==> cbra-0.19.1.tar.gz: cbra-0.19.1/cbra/ext/google/eventarcreceiver.py <==
"""Declares :class:`EventarcReceiver`."""
import asyncio
import typing

import fastapi
import pydantic
from fastapi.responses import Response

import cbra
from cbra.exceptions import AuthenticationRequired
from cbra.exceptions import Forbidden
from cbra.exceptions import Unserializable
from cbra.lib.auth import IPrincipal
from cbra.transaction import Transaction
from .const import GOOGLE_OPENID_SERVER
from .googlejwtprincipal import GoogleJWTPrincipal
from .models import MessagePublished


Message = typing.Union[MessagePublished]


class EventarcReceiver(cbra.BaseRequestHandler, cbra.DependantRunner):
    status_code: int = 202
    message_type: typing.Union = Message
    principal_class: type = GoogleJWTPrincipal

    #: The set of service accounts that are allowed to push messages to this
    #: receiver.
    members: set = set()

    def __init__(self, members: set = None, *args, **kwargs):
        self.members = set(members or [])
        kwargs['method'] = 'POST'
        super().__init__(*args, **kwargs)

    async def authenticate(self,
        token: typing.Optional[IPrincipal],
        audience: str
    ):
        """Authenticate that the request originates from Google."""
        if token is None:
            raise AuthenticationRequired
        token.validate(required={'email', 'sub'})
        if token.get('email') not in self.members:
            self.logger.error(
                "Email address %s is not whitelisted for this endpoint.",
                token.get('email')
            )
            raise Forbidden

    def get_allowed_audience(self, request: fastapi.Request) -> set:
        """Return the set of allowed audiences for this endpoint."""
        assert request.url.scheme is not None # nosec
        assert request.url.netloc is not None # nosec
        assert request.url.path is not None # nosec
        return f"{request.url.scheme}://{request.url.netloc}{request.url.path}"

    def get_correlation_id(self, dto: Message) -> str:
        """Return the correlation id for the message, or ``None``."""
        return None

    def get_handler_args(self, dto: Message, *args, **kwargs) -> tuple:
        return (dto,)

    def get_message(self, dto: Message) -> Message:
        """Hook to parse an object of interest from the message. The default
        implementation simply returns the input object.
        """
        self.logger.info("Received %s from Google Eventarc", type(dto).__name__)
        return dto

    def get_message_handlers(self, dto: Message) -> list:
        """Return the list of handler classes that handle this message."""
        return []

    def get_principal_dependency(self) -> cbra.Dependency:
        """Return the :class:`cbra.Dependency` used to inject the current
        principal.
        """
        return self.principal_class(
            audience=self.get_allowed_audience
        )

    def get_signature_parameters(self):
        return [
            cbra.utils.PositionalArgument(
                name='dto',
                annotation=self.message_type
            ),
        ]

    async def handle(self, dto: Message, *args, **kwargs):
        try:
            messages = self.get_message(dto)
        except pydantic.ValidationError:
            self.logger.error("Invalid message format.")
            return
        except (Unserializable, TypeError, ValueError):
            self.logger.error("Unable to deserialize message.")
            return
        else:
            return await self.run_handlers(
                dto=messages, *args, **kwargs
            )

    async def run_handlers(self,
        request: fastapi.Request,
        dto: Message,
        *args,
        **kwargs
    ):
        """Runs the handlers for the given message."""
        futures = []
        for Handler in self.get_message_handlers(dto):
            futures.append(self._run_handler(request, Handler, dto))
        if futures:
            done, _ = await asyncio.wait(
                futures,
                return_when=asyncio.ALL_COMPLETED
            )
            assert all(map(asyncio.Task.done, done)) # nosec
        self.logger.debug("Ran %s handlers for message %s",
            len(futures), type(dto).__name__)

    async def process_result(self, exception, *args, **kwargs):
        return Response(status_code=202)\
            if not exception\
            else super().process_result(exception=exception, *args, **kwargs)

    async def _run_handler(self, request, handler_class, dto):
        correlation_id = self.get_correlation_id(dto)
        async with Transaction(correlation_id=correlation_id) as tx:
            handler = handler_class(dto, publisher=tx)
            try:
                await self.run_dependant(request, handler)
            except Exception as e:
                self.logger.exception(
                    "Caught fatal %s in handler %s",
                    type(e).__name__,
                    handler_class.__name__
                )

    async def __call__(self,
        request: fastapi.Request,
        response: fastapi.responses.Response,
        principal: IPrincipal,
        *args,
        **kwargs
    ) -> typing.Any:
        """Resolve dependencies and invoke :meth:`handle()`."""
        await self.authenticate(
            token=principal,
            audience=f'{request.url.scheme}://{request.url.netloc}'
        )
        return await super().__call__(request, response, *args, **kwargs)

==> cbra-0.50.0.tar.gz: cbra-0.50.0/cbra/ext/google/eventarcreceiver.py <==
"""Declares :class:`EventarcReceiver`."""
import asyncio
import typing

import fastapi
import pydantic
from fastapi.responses import Response

import cbra
from cbra.ext.ioc import Dependency
from cbra.exceptions import AuthenticationRequired
from cbra.exceptions import Forbidden
from cbra.exceptions import Unserializable
from cbra.lib.auth import IPrincipal
from cbra.transaction import Transaction
from .googlejwtprincipal import GoogleJWTPrincipal
from .models import MessagePublished


Message = typing.Union[MessagePublished]


class EventarcReceiver(cbra.handler.BaseRequestHandler, cbra.DependantRunner):
    status_code: int = 202
    message_type: typing.Union = Message
    principal_class: type = GoogleJWTPrincipal

    #: The set of service accounts that are allowed to push messages to this
    #: receiver.
    members: set = set()

    def __init__(self, members: set = None, *args, **kwargs):
        self.members = set(members or [])
        kwargs['method'] = 'POST'
        super().__init__(*args, **kwargs)

    async def authenticate(self,
        token: typing.Optional[IPrincipal],
        audience: str
    ):
        """Authenticate that the request originates from Google."""
        if token is None:
            raise AuthenticationRequired
        token.validate(required={'email', 'sub'})
        if token.get('email') not in self.members:
            self.logger.error(
                "Email address %s is not whitelisted for this endpoint.",
                token.get('email')
            )
            raise Forbidden

    def get_allowed_audience(self, request: fastapi.Request) -> set: # pragma: no cover
        """Return the set of allowed audiences for this endpoint."""
        assert request.url.scheme is not None # nosec
        assert request.url.netloc is not None # nosec
        assert request.url.path is not None # nosec
        return f"{request.url.scheme}://{request.url.netloc}{request.url.path}"

    def get_correlation_id(self, dto: Message) -> str: # pragma: no cover
        """Return the correlation id for the message, or ``None``."""
        return None

    def get_handler_args(self, dto: Message, *args, **kwargs) -> tuple:
        return (dto,)

    def get_message(self, dto: Message) -> Message: # pragma: no cover
        """Hook to parse an object of interest from the message. The default
        implementation simply returns the input object.
        """
        self.logger.info("Received %s from Google Eventarc", type(dto).__name__)
        return dto

    def get_message_handlers(self, dto: Message) -> list: # pragma: no cover
        """Return the list of handler classes that handle this message."""
        return []

    def get_principal_dependency(self) -> Dependency:
        """Return the :class:`cbra.ext.ioc.Dependency` used to inject the current
        principal.
        """
        return self.principal_class(
            audience=self.get_allowed_audience
        )

    def get_signature_parameters(self):
        return [
            cbra.utils.PositionalArgument(
                name='dto',
                annotation=self.message_type
            ),
        ]

    async def handle(self, dto: Message, *args, **kwargs):
        try:
            messages = self.get_message(dto)
        except pydantic.ValidationError: # pragma: no cover
            self.logger.error("Invalid message format.")
            return
        except (Unserializable, TypeError, ValueError):
            self.logger.error("Unable to deserialize message.")
            return
        else:
            return await self.run_handlers(
                dto=messages, *args, **kwargs
            )

    async def run_handlers(self,
        request: fastapi.Request,
        dto: Message,
        *args,
        **kwargs
    ):
        """Runs the handlers for the given message."""
        futures = []
        for Handler in self.get_message_handlers(dto):
            futures.append(self._run_handler(request, Handler, dto))
        if futures:
            done, _ = await asyncio.wait(
                futures,
                return_when=asyncio.ALL_COMPLETED
            )
            assert all(map(asyncio.Task.done, done)) # nosec
        self.logger.debug("Ran %s handlers for message %s",
            len(futures), type(dto).__name__)

    async def process_result(self, exception, *args, **kwargs):
        return Response(status_code=202)\
            if not exception\
            else super().process_result(exception=exception, *args, **kwargs)

    async def _run_handler(self, request, handler_class, dto):
        correlation_id = self.get_correlation_id(dto)
        async with Transaction(correlation_id=correlation_id) as tx:
            handler = handler_class(dto, publisher=tx)
            try:
                await self.run_dependant(
                    request=request,
                    handle=handler,
                    body=dto.get_object()
                )
            except Exception as e: # pylint: disable=W0703
                self.logger.exception(
                    "Caught fatal %s in handler %s",
                    type(e).__name__,
                    handler_class.__name__
                )

    async def __call__(self,
        request: fastapi.Request,
        response: fastapi.responses.Response,
        principal: IPrincipal,
        *args,
        **kwargs
    ) -> typing.Any:
        """Resolve dependencies and invoke :meth:`handle()`."""
        await self.authenticate(
            token=principal,
            audience=f'{request.url.scheme}://{request.url.netloc}'
        )
        return await super().__call__(request, response, *args, **kwargs)
